"""
MobileMoney API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the MTN MOMO OPEN API.
:copyright: (c) 2022-present rewriteapi
:license: MIT, see LICENSE for more details.
"""

__title__ = 'mobilemoney'
__author__ = 'rewriteapi'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022 rewriteapi'
__version__ = '0.1.2'

from .client import Client
from .errors.errors import *






from .http import HTTPClient
from .collection import Collection
from .disbursements import Disbursements


class Request:
    def __init__(self):
        self.__http = HTTPClient()
        self.__collection = None
        self.__disbursement = None

    def http(self) -> HTTPClient:
        """
        Method to get the HTTP client

        Arguments:
            None
        
        Returns:
            HTTP client : HTTPClient
        """

        return self.__http
    def collection(self, subscription_key: str,  http : HTTPClient) -> Collection:
        """
        Method to get the collection client

        Arguments:
            subscription_key: string
            http: HTTPClient
        
        Returns:
            Collection client : Collection
        """
        self.__collection = Collection(http, subscription_key)
        return self.__collection
    def disbursements(self, subscription_key: str, http: HTTPClient) -> Disbursements:
        """
        Method to get the Disbursements client

        Arguments:
            subscription_key: string
            http: HTTPClient
        
        Returns:
            Disbursements client : Disbursements
        """

        self.__disbursement = Disbursements(http, subscription_key)
        return self.__disbursement



"""
The MIT License (MIT)
Copyright (c) 2022-present rewriteapi
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
from typing import Optional, Any, Dict, TYPE_CHECKING, Union, List, Tuple


from aiohttp import ClientResponse, ClientWebSocketResponse
from requests import Response

_ResponseType = Union[ClientResponse, Response]

__all__ = (
    'MomoException',
    'HTTPException',
    'Unauthorized',
    'MomoServerError',
    'InvalidData',
    'InvalidUniqueIDVersion',
    'Conflict',
    'NotFound',
    'InvalidBasicToken',
    'InvalidBearerToken'
)

def _flatten_error_dict(d: Dict[str, Any], key: str = '') -> Dict[str, str]:
    items: List[Tuple[str, str]] = []
    for k, v in d.items():
        new_key = key + '.' + k if key else k

        if isinstance(v, dict):
            try:
                _errors: List[Dict[str, Any]] = v['_errors']
            except KeyError:
                items.extend(_flatten_error_dict(v, new_key).items())
            else:
                items.append((new_key, ' '.join(x.get('message', '') for x in _errors)))
        else:
            items.append((new_key, v))

    return dict(items)

class MomoException(Exception):
    """Base exception class for MobileMoney.py
    Ideally speaking, this could be caught to handle any exceptions raised from this library.
    """

    pass
class HTTPException(MomoException):
    """Exception that's raised when an HTTP request operation fails.
    Attributes
    ------------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`. In some cases
        this could also be a :class:`requests.Response`.
    text: :class:`str`
        The text of the error. Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
    code: :class:`int`
        The MTN specific error code for the failure.
    """

    def __init__(self, response: _ResponseType, message: Optional[Union[str, Dict[str, Any]]]):
        self.response: _ResponseType = response
        self.status: int = response.status  # type: ignore # This attribute is filled by the library even if using requests
        self.code: int
        self.text: str
        if isinstance(message, dict):
            self.code = message.get('code', 0)
            base = message.get('message', '')
            errors = message.get('errors')
            self._errors: Optional[Dict[str, Any]] = errors
            if errors:
                errors = _flatten_error_dict(errors)
                helpful = '\n'.join('In %s: %s' % t for t in errors.items())
                self.text = base + '\n' + helpful
            else:
                self.text = base
        else:
            self.text = message or ''
            self.code = 0

        fmt = '{0.status} {0.reason} (error code: {1})'
        if len(self.text):
            fmt += ': {2}'

        super().__init__(fmt.format(self.response, self.code, self.text))
    pass

class Unauthorized(HTTPException):
    """Exception that's raised for when status code 401 occurs.
    Subclass of :exc:`HTTPException`
    """
    pass
class MomoServerError(HTTPException):
    """Exception that's raised for when a 500 range status code occurs.
    Subclass of :exc:`HTTPException`.
    """
    pass
class InvalidData(HTTPException):
    """Exception that's raised for when a 400 range status code occurs.
    Subclass of :exc:`HTTPException`.
    """
    pass
class NotFound(HTTPException):
    """Exception that's raised for when a 404 range status code occurs.
    Subclass of :exc:`HTTPException`.
    """
    pass
class Conflict(HTTPException):
    """Exception that's raised for when a 409 range status code occurs.
    Subclass of :exc:`HTTPException`.
    """
    pass
class InvalidUniqueIDVersion(Exception):
    """Exception that's raised for when UUID is not a valid version 4.
    Subclass of :exc:`Exception`
    """
    pass
class InvalidBasicToken(Exception):
    """Exception that's raised for a invalid Basic token type.
    Subclass of :exc:`Exception`
    """
    pass
class InvalidBearerToken(Exception):
    """Exception that's raised for a invalid Bearer token type.
    Subclass of :exc:`Exception`
    """
    pass


"""
The MIT License (MIT)
Copyright (c) 2022-present rewriteapi
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from .http import HTTPClient
from .route import Route
from typing import Dict, Optional, Tuple
from ..utils.utils import is_valid_bearer_token, is_valid_basic_token, COLLECTION_PATH, errors_manager, is_valid_id_4
from ..errors.errors import InvalidBasicToken, InvalidBearerToken, InvalidUniqueIDVersion

class Collection:
    """
    Represent Collection Client used by collection user

    Arguments:
        http: a HTTP client class
        subscription_key: a string

    Returns:
        None
    """

    def __init__(self, http: HTTPClient,  subscription_key: str)->None:
        self.subscription_key = subscription_key
        self.http = http

    async def create_access_token(self, authorization: str)-> Tuple:
        """
        Method to create access token for collection user

        Arguments:
            authorization: a string

        Return:
            Tuple : (boolean, data)
        """

        if is_valid_basic_token(authorization):

            headers = {'Authorization':authorization, 'Ocp-Apim-Subscription-Key':self.subscription_key}
            
            response = await self.http.request(Route('POST', 
            COLLECTION_PATH['create_access_token'][Route.ENV[self.http.isLive]],
            self.http.isLive,
            headers))

            if response.status == 200:
                #want to return a true cause anything can happen, just for a check purpose
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBasicToken('Invalid Basic Token Type given')

    async def get_account_balance(self, authorization: str, target: str)->Tuple:
        """
        Method to get balance for collection user

        Arguments:
            authorization: string
            target: string

        Return
            Tuple : (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET', 
                COLLECTION_PATH['get_account_balance'][Route.ENV[self.http.isLive]],
                self.http.isLive,
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def get_account_balance_in(self, currency: str, authorization: str, target: str)->Tuple:
        """
        Method to get balance in specific currency for collection user

        Arguments:
            currency: string
            authorization: string
            target: string

        Return:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET', 
                COLLECTION_PATH['get_account_balance_in'][Route.ENV[self.http.isLive]].format(currency=currency),
                self.http.isLive,
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
            
        
    async def get_basic_user_info(self, msisdn: str, authorization: str, target: str)-> Tuple:
        """
        Method to get basic user info without consent for collection user

        Arguments:
            msisdn: string
            authorization: string
            target: string
            
        Retruns:
            Tuple : (boolean, data)
        """
        if is_valid_bearer_token(authorization):
            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET',
                COLLECTION_PATH['get_basic_info'][Route.ENV[self.http.isLive]].format(MSISDN=msisdn),
                self.http.isLive,
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def ask_user_info(self, authorization: str, target: str) -> Tuple:
        """
        Method to get user info with consent for collection user

        Arguments:
            authorization: string
            target : string
        
        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):
            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET',
                COLLECTION_PATH['get_user_info'][Route.ENV[self.http.isLive]],
                self.http.isLive, 
                headers))
            
            if response.status == 200:
                return(True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def request_to_pay(self, authorization: str, uuid: str, target: str, body: Dict, callback: Optional[str]=None) -> Tuple:
        """
        Method to request a payment for collection user

        Arguments:
            authorization: string
            uuid: string
            target: string
            body: dictionary
            callback [optional]: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            #check if uuid is valid and raise error if not
            self.http.uuid_checker(uuid)
            
            #callback url is optional, check if parsed and replace the key to avoid any error if not
            callkey = 'X-Callback-Url'
            if callback is None:
                callback = "No callback url given"
                callkey = "No-callback-Url"

            headers = {'Authorization':authorization, 
                'X-Target-Environment':target, 
                'Ocp-Apim-Subscription-Key':self.subscription_key,
                'X-Reference-Id':uuid,
                'Content-Type':'application/json',
                callkey:callback
                }
            
            response = await self.http.request(Route('POST', 
                COLLECTION_PATH['request_to_pay'][Route.ENV[self.http.isLive]],
                self.http.isLive, 
                headers, 
                body))

            if response.status == 202:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
    
    async def get_withdraw_status(self, authorization: str, uuid: str, target : str = 'sandbox') -> Tuple:
        """
        Method to get a withdrawal status for collection user

        Arguments:
            authorization: string
            uuid: string
            target: string
        
        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):
            
            self.http.uuid_checker(uuid)

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET', 
                COLLECTION_PATH['withdraw_status'][Route.ENV[self.http.is_Live]].format(referenceId=uuid),
                self.http.is_Live,
                headers))
            
            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
    

    async def withdraw(self, authorization: str, uuid: str, target: str, body: Dict, callback: Optional[str]=None) -> Tuple:
        """
        Method to withdraw money for collection user

        Arguments:
            authorization: string
            uuid: string
            target: srting
            body: dictionary
            callback [optional]: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            #check if uuid is valid and raise error if not
            self.http.uuid_checker(uuid)
            
            #callback url is optional, check if parsed and replace the key to avoid any error if not
            callkey = 'X-Callback-Url'
            if callback is None:
                callback = "No callback url given"
                callkey = "No-callback-Url"

            headers = {'Authorization':authorization, 
                'X-Target-Environment':target, 
                'Ocp-Apim-Subscription-Key':self.subscription_key,
                'X-Reference-Id':uuid,
                'Content-Type':'application/json',
                callkey:callback
                }
            
            response = await self.http.request(Route('POST', 
                COLLECTION_PATH['withdraw'][Route.ENV[self.http.isLive]],
                self.http.isLive, 
                headers, 
                body))

            if response.status == 202:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')

    async def isActive(self, account: str, account_type: Optional[str]='msisdn', authorization : Optional[str]=None, target : str = 'sandbox') -> Tuple:
        """
        Method to check if an account is active for a collection user

        Arguments:
            account: string
            account_type [optional default set to 'msisdn']: string
            authorization: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """

        headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

        response = await self.http.request(Route('GET',
            COLLECTION_PATH['is_active'][Route.ENV[self.http.isLive]].format(accountHolderIdType=account_type, accountHolderId=account),
            self.http.is_Live,
            headers))
        
        if response.status == 200:
            return (True, self.http.data)
        else:
            errors_manager(response, self.http.data)

    


"""
The MIT License (MIT)
Copyright (c) 2022-present rewriteapi
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import asyncio
import datetime
import logging
import aiohttp
import json
from typing import ClassVar, Tuple, Union, Dict, Any, Optional
from ..utils import utils
from ..errors.errors import Conflict, MomoException, HTTPException, Unauthorized, MomoServerError, InvalidData, InvalidUniqueIDVersion
from .route import Route
"""
Note : Authorization is api user ID and api key
"""




class HTTPClient:
    """Represent the HTTP client"""

    def __init__(
        self,
        connector: Optional[aiohttp.BaseConnector] = None,
        ) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.connector: aiohttp.BaseConnector = connector,
        self.__session: aiohttp.ClientSession = None
        user_agent =  'MobileMoney python version'
        self.user_agent = user_agent
        self.isLogged = False
        self.isLive = True
        self.data: Optional[Union[Dict[str, Any], str]] = None

    async def login(self)-> None:
        if not self.isLogged:
            self.__session = aiohttp.ClientSession()
            self.isLogged = True
            
            
    async def logout(self)-> None:
        if self.isLogged:
            await self.__session.close()
            self.isLogged = False 
            

    def is_sandbox(self)-> None:
        """Function to turn the library from live to sandbox environment
        Caution : Make sure to activate this before any request if you are not in live
        """
        self.isLive = False

    def uuid_checker(self, uuid: str)-> None:
        """
        Function in charge of checking UUID version and raise error if False
        :parameter: uuid
        """
        if not utils.is_valid_id_4(uuid):
            raise InvalidUniqueIDVersion("Invalid UUID version 4")


    async def request(
        self,
        route: Route    
        )-> aiohttp.ClientResponse:

        await self.login()
        method = route.method
        url = route.url

        #creating headers and adding our user agent for potential statistic or analyse later
        #hope MTN store it lmao
        route.headers['User-Agent'] = self.user_agent

        #response: Optional[aiohttp.ClientResponse] = None
        
        try:
            if route.body is not None:
                body = json.dumps(route.body[0])  
            else:
                body = json.dumps(route.body)
        except KeyError:
            body = json.dumps(route.body)
        
        try:
            async with self.__session.request(method, url, data=body, headers=route.headers) as response:
                self.data = await utils.json_or_text(response)
                await self.logout()
                return response


        except Exception as e:
            print(f'Something wrong when sending request to mtn : {e}')
        
        await self.logout()

    async def create_api_user(self, uuid: str, subscription_key: str, url_callback : Optional[str] = None)->bool:
        """
        Function in charge of creating API USER from wallet provider
        :parameter: uuid, subscription_key
        :return: True or False 
        :return type: bool
        """


        self.uuid_checker(uuid)

        headers = {
            "X-Reference-Id":uuid,
            "Ocp-Apim-Subscription-Key":subscription_key,
            "Content-Type":"application/json"
        }

        if url_callback is None :
            url_callback = 'string'

        body = {
            "providerCallbackHost":url_callback
        }
        
        response: Optional[aiohttp.ClientResponse] = None
        response = await self.request(Route('POST',
        utils.PATH['create_apiuser'][Route.ENV[self.isLive]], 
        self.isLive,
        headers, 
        body))

        #if done 
        if response.status == 201:
            return True
        else:
            utils.errors_manager(response, self.data)

    async def get_api_user(self, uuid: str, subscription_key: str) -> Tuple:
        """
        Function in charge of getting API USER from wallet provider
        :parameter: uuid, subscription_key
        :return: (boolean, user)
        :return type: Tuple
        """
        self.uuid_checker(uuid)

        headers = {
            "Ocp-Apim-Subscription-Key":subscription_key
        }
        response: Optional[aiohttp.ClientResponse] = None
        response =  await self.request(Route('GET', 
        utils.PATH['get_apiuser'][Route.ENV[self.isLive]].format(uuid=uuid), 
        self.isLive,
        headers
        ))

        #if done
        if response.status == 200:
            return (True, self.data)
        else:
            utils.errors_manager(response, self.data)

    async def create_api_key(self, uuid: str, subscription_key: str) -> Tuple:
        """
        Function in charge of creating API KEY from wallet provider
        :parameter: uuid, subscription_key
        :return:(boolean, API KEY)
        :return type: Tuple
        """
        self.uuid_checker(uuid)

        headers = {
            "Ocp-Apim-Subscription-Key":subscription_key
        }
        response: Optional[aiohttp.ClientResponse] = None
        response =  await self.request(Route('POST', 
        utils.PATH['create_apikey'][Route.ENV[self.isLive]].format(apiuser=uuid),
        self.isLive, 
        headers
        ))
        if response.status == 201:
            return (True, self.data)
        else:
            utils.errors_manager(response, self.data)
    






"""
The MIT License (MIT)
Copyright (c) 2022-present rewriteapi
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import datetime
import logging
from typing import Any, Optional, Dict, Tuple
import aiohttp


from .request.request import Request
from .utils.utils import get_reference_id, b64_encode
class Client:
    """
    client agent
    """

    def __init__(self) -> None:
        self.request = Request()
        self.http = self.request.http()

    def collection(self, subscription_key: str) -> Any:
        """
        Method to get the collection client

        Arguments:
            subscription_key: string
        
        Returns:
            Collection client : Collection
        """
        return self.request.collection(subscription_key, self.http)

    def disbursements(self, subscription_key: str)-> Any:
        """
        Method to get the Disbursements client

        Arguments:
            subscription_key: string
        
        Returns:
            Disbursements client : Disbursements
        """
        return self.request.disbursements(subscription_key, self.http)

    async def create_api_user(self, uuid: str, subscription_key: str, url_callback : Optional[str] = None)->bool:
        """
        Function in charge of creating API USER from wallet provider
        :parameter: uuid, subscription_key, url_callback [optional]
        :return: True or False 
        :return type: bool
        """
        return await self.http.create_api_user(uuid, subscription_key,url_callback)

    async def get_api_user(self, uuid: str, subscription_key: str) -> Tuple:
        """
        Function in charge of getting API USER from wallet provider
        :parameter: uuid, subscription_key
        :return: (boolean, user)
        :return type: Tuple
        """
        return await self.http.get_api_user(uuid, subscription_key)

    async def create_api_key(self, uuid: str, subscription_key: str) -> Tuple:
        """
        Function in charge of creating API KEY from wallet provider
        :parameter: uuid, subscription_key
        :return:(boolean, API KEY)
        :return type: Tuple
        """
        return await self.http.create_api_key(uuid, subscription_key)

    def get_reference_id(self)-> str:
        """
        Function to create UUID version 4
        Arguments:
            None

        Returns:
            UUID: string
        """
        return get_reference_id()

    def is_sandbox(self) -> None:
        """method to turn on sandbpx environement

        CAUTION
        -------------------
        Make sure to call this function firstly before any request or you will make request to live environment
        """
        self.http.is_sandbox()

    def basic_token(self, apiuser: str, apikey: str) -> str:
        """Method to create basic token from API USER & API KEY"""

        data =  f'{apiuser}:{apikey}'
        encoded = b64_encode(data)
        return f"Basic {encoded}"

    def bearer_token(self, token: str)-> str:
        """Method to convert access token to Bearer token"""

        return f"Bearer {token}"



"""
The MIT License (MIT)
Copyright (c) 2022-present rewriteapi
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import json
from typing import Any, Dict, Union, TYPE_CHECKING, Optional
import base64
from urllib.parse import urlencode
import uuid
import aiohttp
from ..errors.errors import *



from aiohttp import ClientResponse, ClientWebSocketResponse
from requests import Response

_ResponseType = Union[ClientResponse, Response]

def encode_params(params: Any):
    
    if type(params) is not Dict:
        params = Dict(params)
    
    return urlencode(params)

def _to_json(obj: Any) -> str:
    return json.dumps(obj, separators=(',', ':'), ensure_ascii=True)

_from_json = json.loads

def b64_encode(data: str)->str:
    """This function encode data to base64"""

    data_bytes = data.encode('ascii')
    base64_bytes = base64.b64encode(data_bytes)
    data_encoded = base64_bytes.decode('ascii')

    return data_encoded

def is_valid_bearer_token(token: str)->bool:
    if token.startswith('Bearer '):
        return True
    else:
        return False

def is_valid_basic_token(token: str)->bool:
    if token.startswith('Basic '):
        return True
    else:
        return False

async def json_or_text(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    text = await response.text(encoding='utf-8')

    try:
        if response.headers['content-type'] == 'application/json;charset=utf-8' or response.headers['content-type'] == 'application/json':
            return _from_json(text)
    except KeyError:
        pass

    return text

def get_reference_id()->str:
    """A method helping api user to get X-Reference-id
        Arguement : None
        Return : UUID 
        Return type : str
    """
    unique_id = uuid.uuid4()
    return str(unique_id)

def is_valid_id_4(unique_id: str)->bool:
    """A method checking if X-Reference-id is a valid UUID version 4
        parameter : X-Reference-id
        Return type : bool
    """
    try:
        id_convert_test = uuid.UUID(unique_id, version=4)
    except ValueError:
        return False
    return str(id_convert_test) == unique_id

def errors_manager(response: _ResponseType, data: Optional[Union[str, Dict[str, Any]]] = "")-> None:

    print(response)
    print(data)

    if response.status == 400:
        raise InvalidData(response, data)
    elif response.status == 409:
        raise Conflict(response, data)
    elif response.status == 500:
        raise MomoServerError(response, data)
    elif response.status == 401:
        raise Unauthorized(response, data)
    elif response.status == 404:
        raise NotFound(response, data)
    else:
        raise HTTPException(response, data)

PATH = {
    "create_apiuser":{
        "sandbox":"/v1_0/apiuser",
        "live":"/provisioning/v1_0/apiuser"
    },
    "get_apiuser":{
        "sandbox":"/v1_0/apiuser/{uuid}",
        "live":""
    },
    "create_apikey":{
        "sandbox":"/v1_0/apiuser/{apiuser}/apikey",
        "live":""
    }
}

COLLECTION_PATH = {
    "create_access_token":{
        "sandbox":"/collection/token/",
        "live":"/collection/token/"
    },
    "get_account_balance":{
        "sandbox":"/collection/v1_0/account/balance",
        "live":"/collection/v1_0/account/balance"
    },
    "get_account_balance_in":{
        "sandbox":"/collection/v1_0/account/balance/{currency}",
        "live":"/collection/v1_0/account/balance/{currency}"
    },
    "get_basic_info":{
        "sandbox":"/collection/v1_0/accountholder/msisdn/{MSISDN}/basicuserinfo",
        "live":"/collection/v1_0/accountholder/msisdn/{MSISDN}/basicuserinfo"
    },
    "get_user_info":{
        "sandbox":"/collection/oauth2/v1_0/userinfo",
        "live":"/collection/oauth2/v1_0/userinfo"
    },
    "request_to_pay":{
        "sandbox":"/collection/v1_0/requesttopay",
        "live":"/collection/v1_0/requesttopay"
    },
    "withdraw":{
        "sandbox":"/collection/v2_0/requesttowithdraw",
        "live":"/collection/v2_0/requesttowithdraw"
    },
    "withdraw_status":{
        "sandbox":"/collection/v1_0/requesttowithdraw/{referenceId}",
        "live":"/collection/v1_0/requesttowithdraw/{referenceId}"
    },
    "is_active":{
        "sandbox":"/collection/v1_0/accountholder/{accountHolderIdType}/{accountHolderId}/active",
        "live":"/collection/v1_0/accountholder/{accountHolderIdType}/{accountHolderId}/active"
    }
}

DISBURSEMENTS_PATH = {
    "create_access_token":{
        "sandbox":"/disbursement/token/",
        "live":"/disbursement/token/"
    },
    "get_account_balance":{
        "sandbox":"/disbursement/v1_0/account/balance",
        "live":"/disbursement/v1_0/account/balance"
    },
    "get_account_balance_in":{
        "sandbox":"/disbursement/v1_0/account/balance/{currency}",
        "live":"/disbursement/v1_0/account/balance/{currency}"
    },
    "get_basic_info":{
        "sandbox":"/disbursement/v1_0/accountholder/msisdn/{MSISDN}/basicuserinfo",
        "live":"/disbursement/v1_0/accountholder/msisdn/{MSISDN}/basicuserinfo"
    },
    "get_user_info":{
        "sandbox":"/disbursement/oauth2/v1_0/userinfo",
        "live":"/disbursement/oauth2/v1_0/userinfo"
    },
    "is_active":{
        "sandbox":"/disbursement/v1_0/accountholder/{accountHolderIdType}/{accountHolderId}/active",
        "live":"/disbursement/v1_0/accountholder/{accountHolderIdType}/{accountHolderId}/active"
    },
    "get_transfer_status":{
        "sandbox":"/disbursement/v1_0/transfer/{referenceId}",
        "live":"/disbursement/v1_0/transfer/{referenceId}"
    },
    "deposit":{
        "sandbox":"/disbursement/v2_0/deposit",
        "live":"/disbursement/v2_0/deposit"
    },
    "transfer":{
        "sandbox":"/disbursement/v1_0/transfer",
        "live":"/disbursement/v1_0/transfer"
    },
    "refund":{
        "sandbox":"/disbursement/v2_0/refund",
        "live":"/disbursement/v2_0/refund"
    },
    "get_transfer_status":{
        "sandbox":"/disbursement/v1_0/transfer/{referenceId}",
        "live":"/disbursement/v1_0/transfer/{referenceId}"
    },
    "get_refund_status":{
        "sandbox":"/disbursement/v1_0/refund/{referenceId}",
        "live":"/disbursement/v1_0/refund/{referenceId}"
    }
}


"""
The MIT License (MIT)
Copyright (c) 2022-present rewriteapi
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from typing import Dict, Tuple, Optional

from ..errors.errors import InvalidBasicToken, InvalidBearerToken

from ..utils.utils import errors_manager, is_valid_basic_token, is_valid_bearer_token, DISBURSEMENTS_PATH
from .http import HTTPClient
from .route import Route


class Disbursements:
    """
    Represent Disbursement Client used by disbursement user

    Arguments:
        http: a HTTP client class
        subscription_key: a string

    Returns:
        None
    """
    def __init__(self, http: HTTPClient, subscripyion_key: str)->None:
        self.http = http
        self.subscripyion_key = subscripyion_key

    async def create_access_token(self, authorization: str) -> Tuple:
        """
        Method to create access token for Disbursements user

        Arguments:
            authorization: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_basic_token(authorization):
            headers = {"Authorization":authorization, "Ocp-Apim-Subscription-Key":self.subscripyion_key}
            response = await self.http.request(Route('POST', 
                DISBURSEMENTS_PATH['create_access_token'][Route.ENV[self.http.isLive]],
                self.http.isLive, 
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBasicToken('Invalid Basic Token Type given')

    async def get_account_balance(self, authorization: str, target: str)->Tuple:
        """
        Method to get account balance for Disbursement user

        Arguments:
            authorization: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """
        if is_valid_bearer_token(authorization):

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET', 
                DISBURSEMENTS_PATH['get_account_balance'][Route.ENV[self.http.isLive]],
                self.http.isLive,
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def get_account_balance_in(self, currency: str, authorization: str, target: str)->Tuple:
        """
        Method to get balance in specific currency for disbursement user

        Arguements:
            currency: string
            authorization: string
            target:string
        
        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET', 
                DISBURSEMENTS_PATH['get_account_balance_in'][Route.ENV[self.http.isLive]].format(currency=currency),
                self.http.isLive,
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def get_basic_user_info(self, msisdn: str, authorization: str, target: str)-> Tuple:
        """
        Method to get basic user info without consent for disbursement user

        Arguments:
            msisdn: string
            authorization: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """
        if is_valid_bearer_token(authorization):
            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET',
                DISBURSEMENTS_PATH['get_basic_info'][Route.ENV[self.http.isLive]].format(MSISDN=msisdn),
                self.http.isLive,
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def ask_user_info(self, authorization: str, target: str) -> Tuple:
        """
        Method to get user info with consent for disbursement user

        Arguments:
            authorization: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):
            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET',
                DISBURSEMENTS_PATH['get_user_info'][Route.ENV[self.http.isLive]],
                self.http.isLive, 
                headers))
            
            if response.status == 200:
                return(True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def get_deposit_status(self, uuid: str, authorization: str, target: str) -> Tuple:
        """
        Method to get transfer status for disbursement user

        Arguments:
            uuid: string
            authorization: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            self.http.uuid_checker(uuid)

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET', 
                DISBURSEMENTS_PATH['get_tranfer_status'][Route.ENV[self.http.isLive]].format(referenceId=uuid),
                self.http.isLive, 
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            InvalidBearerToken('Invalid Bearer Token Type given')
         
    async def deposit(self, uuid: str, authorization: str, target:str, body: Dict, url_callback: Optional[str]= None) -> Tuple:
        """
        Method to deposit for disbursement user 

        Arguments:
            uuid: string
            authorization: string
            target: string
            body: dictionary
            url_callback [optional]: string

        Returns:
            Tuple: (boolean, data)
        """
        
        if is_valid_bearer_token(authorization):

            self.http.uuid_checker(uuid)

            callkey = 'X-Callback-Url'
            if url_callback is None:
                url_callback = "No callback url given"
                callkey = "No-callback-Url"

            headers = {
                'Authorization':authorization, 
                'X-Target-Environment':target, 
                'Ocp-Apim-Subscription-Key':self.subscription_key,
                'X-Reference-Id':uuid,
                'Content-Type':'application/json',
                callkey:url_callback
            }

            response = await self.http.request(Route('POST', 
                DISBURSEMENTS_PATH['deposit'][Route.ENV[self.http.isLive]],
                headers, 
                body))
            
            if response.status == 202:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)

        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def transfer(self, uuid: str, authorization: str, target:str, body: Dict, url_callback: Optional[str]= None) -> Tuple:
        """
        Method to transfer for disbursement user 

        Arguments:
            uuid: string
            authorization: string
            target: string
            body: dictionary
            url_callback [optional]: string

        Returns:
            Tuple: (boolean, data)
        """
        
        if is_valid_bearer_token(authorization):

            self.http.uuid_checker(uuid)

            callkey = 'X-Callback-Url'
            if url_callback is None:
                url_callback = "No callback url given"
                callkey = "No-callback-Url"

            headers = {
                'Authorization':authorization, 
                'X-Target-Environment':target, 
                'Ocp-Apim-Subscription-Key':self.subscription_key,
                'X-Reference-Id':uuid,
                'Content-Type':'application/json',
                callkey:url_callback
            }

            response = await self.http.request(Route('POST', 
                DISBURSEMENTS_PATH['transfer'][Route.ENV[self.http.isLive]],
                headers, 
                body))
            
            if response.status == 202:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)

        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def refund(self, uuid: str, authorization: str, target:str, body: Dict, url_callback: Optional[str]= None) -> Tuple:

        """
        Method to transfer for disbursement user 

        Arguements:
            uuid: string
            authorization: string
            target: string
            body: dictionary
            url_callback [optional]: string

        Returns:
            Tuple: (boolean, data)
        """
        
        if is_valid_bearer_token(authorization):

            self.http.uuid_checker(uuid)

            callkey = 'X-Callback-Url'
            if url_callback is None:
                url_callback = "No callback url given"
                callkey = "No-callback-Url"

            headers = {
                'Authorization':authorization, 
                'X-Target-Environment':target, 
                'Ocp-Apim-Subscription-Key':self.subscription_key,
                'X-Reference-Id':uuid,
                'Content-Type':'application/json',
                callkey:url_callback
            }

            response = await self.http.request(Route('POST', 
                DISBURSEMENTS_PATH['refund'][Route.ENV[self.http.isLive]],
                headers, 
                body))
            
            if response.status == 202:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)

        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def get_transfer_status(self, uuid: str, authorization: str, target: str) -> Tuple:
        """
        Method to get transfer status for disbursement user

        Arguments:
            uuid: string
            authorization: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            self.http.uuid_checker(uuid)

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}
            
            response = await self.http.request(Route('GET',
                DISBURSEMENTS_PATH['get_transfer_status'][Route.ENV[self.http.isLive]].format(referenceId=uuid),
                self.http.isLive, 
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
     
    async def get_refund_status(self, uuid: str, authorization: str, target: str) -> Tuple:
        """
        Method to get transfer status for disbursement user

        Arguments:
            uuid: string
            authorization: string
            target: string
        
        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):

            self.http.uuid_checker(uuid)

            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}
            
            response = await self.http.request(Route('GET',
                DISBURSEMENTS_PATH['get_refund_status'][Route.ENV[self.http.isLive]].format(referenceId=uuid),
                self.http.isLive, 
                headers))

            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        
    async def isActive(self, account: str, authorization : str, account_type: Optional[str]='msisdn',  target : str = 'sandbox') -> Tuple:
        """
        Method to check if an account is active for a disbursement user

        Arguements:
            account: string
            authorization: string
            account_type [optional default set to 'msisdn']: string
            target: string

        Returns:
            Tuple: (boolean, data)
        """

        if is_valid_bearer_token(authorization):
            headers = {'Authorization':authorization, 'X-Target-Environment':target, 'Ocp-Apim-Subscription-Key':self.subscription_key}

            response = await self.http.request(Route('GET',
                DISBURSEMENTS_PATH['is_active'][Route.ENV[self.http.isLive]].format(accountHolderIdType=account_type, accountHolderId=account),
                self.http.is_Live,
                headers))
            
            if response.status == 200:
                return (True, self.http.data)
            else:
                errors_manager(response, self.http.data)
        else:
            raise InvalidBearerToken('Invalid Bearer Token Type given')
        



from setuptools import setup

requirements = []
with open('requirements.txt') as f:
  requirements = f.read().splitlines()

readme = ''
with open('README.rst') as f:
    readme = f.read()

packages = [
  'mobilemoney',
  'mobilemoney.request',
  'mobilemoney.errors',
  'mobilemoney.utils'
  ]

setup(name='mobilemoney.py',
      author='rewriteapi',
      url='https://rewriteapi.cm/momo',
      author_email="<support@rewriteapi.cm>",
      project_urls={
        "Documentation": "https://momopy.rewriteapi.cm",
        "Website":"https://rewriteapi.cm",
        "Repository":"https://github.com/rewriteapi/mobilemoney.py"
      },
      version='0.1.2',
      packages=packages,
      long_description=readme,
      long_description_content_type="text/markdown",
      license='MIT',
      description='A Python wrapper for the MoMo Open API',
      include_package_data=True,
      install_requires=requirements,
      keywords=['python', 'mobilemoney', 'MTN Money', 'rewriteapi', 'MTN API', 'mobilemoney-py'],
      python_requires='>=3.8.0',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Typing :: Typed',
      ]

)